"""Module for testing utility functions."""


import numpy as np
import pandas as pd


def is_empty(value, empty_values = ('', [], None, np.nan)):
    """Whether an object is empty or not from scraping point of view."""
    if type(value) == str:
        value = value.strip()
    if type(value) == list:
        if len(value) == 0:
            return True
        else:
            return all(is_empty(el) for el in value)
    elif type(value) == dict:
        if len(value) == 0:
            return True
        else:
            return all(is_empty(el) for el in value.values())
    else:
        return value in empty_values


def calc_field_emptiness(df, empty_values=('', [], None, np.nan)):
    """Calculate emptiness of DataFrame columns."""
    stats = []
    for col in df:
        empty = df[col].apply(is_empty, empty_values=empty_values)
        n_empty = empty.sum()
        n_total = len(empty)
        empty_pct = n_empty / n_total * 100
        stats.append({
            'field': col,
            'n_empty': n_empty,
            'n_total': n_total,
            'empty_pct': round(empty_pct, 1)
        })
    stats_df = pd.DataFrame(stats).sort_values('empty_pct', ascending=False)
    return stats_df



"""Module for OikotieApartment spider."""


import time

from functools import partial

from scrapy import Item, Field, Selector, Request
from scrapy.crawler import Spider
from scrapy.exceptions import DropItem, CloseSpider
from scrapy.http import HtmlResponse
from scrapy.linkextractors import LinkExtractor
from scrapy.loader import ItemLoader
from scrapy.loader.processors import TakeFirst, Identity, MapCompose, \
    Compose, Join

from finscraper.http import SeleniumCallbackRequest
from finscraper.scrapy_spiders.mixins import FollowAndParseItemMixin
from finscraper.utils import strip_join, safe_cast_int, strip_elements, \
    drop_empty_elements


class _OikotieApartmentSpider(Spider):
    name = 'oikotieapartment'
    start_urls = ['https://asunnot.oikotie.fi/myytavat-asunnot']
    follow_link_extractor = LinkExtractor(
        allow_domains=('asunnot.oikotie.fi'),
        allow=(r'.*\/myytavat-asunnot\/.*'),
        deny=('.*?origin\=.*'),
        deny_domains=(),
        canonicalize=True
    )
    item_link_extractor = LinkExtractor(
        allow_domains=('asunnot.oikotie.fi'),
        allow=(rf'.*/myytavat-asunnot/.*/[0-9]+'),
        deny=('.*?origin\=.*'),
        deny_domains=(),
        canonicalize=True
    )
    custom_settings = {
        'ROBOTSTXT_OBEY': False,  # No robots.txt, will fail with yes
        'DOWNLOADER_MIDDLEWARES': {
            'finscraper.middlewares.SeleniumCallbackMiddleware': 800
        },
        'DEFAULT_REQUEST_HEADERS': {
            'Connection': 'keep-alive',
            'Cache-Control': 'max-age=0',
            'DNT': '1',
            'Upgrade-Insecure-Requests': '1',
            'User-Agent': 'Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/78.0.3904.87 Safari/537.36',
            'Sec-Fetch-User': '?1',
            'Accept': 'text/html,application/xhtml+xml,application/xml;q=0.9,image/webp,image/apng,*/*;q=0.8,application/signed-exchange;v=b3',
            'Sec-Fetch-Site': 'same-origin',
            'Sec-Fetch-Mode': 'navigate',
            'Accept-Encoding': 'gzip, deflate, br',
            'Accept-Language': 'en-US,en;q=0.9',
        }
    }
    itemcount = 0
    title2field = {
        # Perustiedot
        'Sijainti': 'location',
        'Kaupunki': 'city',
        'Kaupunginosa': 'district',
        'Kohdenumero': 'oikotie_id',
        'Kerros': 'floor',
        'Asuinpinta-ala': 'life_sq',
        'Tontin pinta-ala': 'property_sq',
        'Kokonaispinta-ala': 'total_sq',
        'Huoneiston kokoonpano': 'room_info',
        'Huoneita': 'number_of_rooms',
        'Kunto': 'condition',
        'Kunnon lisätiedot': 'condition_details',
        'Lisätietoa vapautumisesta': 'availability',
        'Keittiön varusteet': 'kitchen_appliances',
        'Kylpyhuoneen varusteet': 'bathroom_appliances',
        'Ikkunoiden suunta': 'window_direction',
        'Parveke': 'has_balcony',
        'Parvekkeen lisätiedot': 'balcony_details',
        'Säilytystilat': 'storage_space',
        'Näkymät': 'view',
        'Tulevat remontit': 'future_renovations',
        'Tehdyt remontit': 'completed_renovations',
        'Asunnossa sauna': 'has_sauna',
        'Saunan lisätiedot': 'sauna_details',
        'Asumistyyppi': 'housing_type',
        'Palvelut': 'services',
        'Lisätiedot': 'additional_info',
        'Kiinteistötunnus': 'property_id',
        'Kohde on': 'apartment_is',
        'Tietoliikennepalvelut': 'telecommunication_services',

        # Hintatiedot ja muut kustannukset
        'Velaton hinta': 'price_no_tax',
        'Myyntihinta': 'sales_price',
        'Lainaosuuden maksu': 'shared_loan_payment',
        'Neliöhinta': 'price_per_sq',
        'Velkaosuus': 'share_of_liabilities',
        'Kiinnitykset': 'mortgages',
        
        'Rahoitusvastike': 'financial_charge',
        'Hoitovastike': 'condominium_payment',
        'Yhtiövastike': 'maintenance_charge',
        'Vesimaksu': 'water_charge',
        'Vesimaksun lisätiedot': 'water_charge_details',
        'Lämmityskustannukset': 'heating_charge',
        'Muut kustannukset': 'other_costs',

        # Talon ja tontin tiedot
        'Uudiskohde': 'is_brand_new',
        'Taloyhtiön nimi': 'housing_company_name',
        'Rakennuksen tyyppi': 'building_type',
        'Rakennusvuosi': 'build_year',
        'Rakennusvuoden lisätiedot': 'build_year_details',
        'Huoneistojen lukumäärä': 'number_of_apartments',
        'Kerroksia': 'total_floors',
        'Hissi': 'building_has_elevator',
        'Taloyhtiössä on sauna': 'building_has_sauna',
        'Rakennusmateriaali': 'building_material',
        'Kattotyyppi': 'roof_type',
        'Energialuokka': 'energy_class',
        'Energiatoditus': 'has_energy_certificate',
        'Kiinteistön antennijärjestelmä': 'antenna_system',
        'Tontin koko': 'property_size',
        'Kiinteistönhoito': 'maintenance',
        'Isännöinti': 'real_estate_management',
        'Kaavatiedot': 'plan_info',
        'Kaavatilanne': 'plan',
        'Liikenneyhteydet': 'traffic_communication',
        'Lämmitys': 'heating',

        # Tilat ja materiaalit
        'Pysäköintitilan kuvaus': 'parking_space_description',
        'Yhteiset tilat': 'common_spaces',
        'Pintamateriaalit': 'wallcovering'
    }

    def __init__(self, *args, **kwargs):
        """Fetch oikotie.fi apartments.
        
        Args:
        """
        kwargs['follow_request_type'] = SeleniumCallbackRequest
        super(_OikotieApartmentSpider, self).__init__(*args, **kwargs)

    def start_requests(self):
        for url in self.start_urls:
            yield SeleniumCallbackRequest(
                url, selenium_callback=self._handle_start)

    @staticmethod
    def _handle_start(request, spider, driver):
        driver.get(request.url)
        driver.find_element_by_xpath(
            '//div[contains(@class, "sccm-button-green")]').click()
        return HtmlResponse(
            driver.current_url,
            body=driver.page_source.encode('utf-8'),
            encoding='utf-8',
            request=request
        )

    def parse(self, resp, to_parse=False):
        """Parse items and follow links based on defined link extractors."""
        if (self.itemcount and 
            self.itemcount == self.settings.get('CLOSESPIDER_ITEMCOUNT', 0)):
                raise CloseSpider

        if to_parse:
            yield self._parse_item(resp)
            self.itemcount += 1

        # Parse items and further on extract links from those pages
        item_links = self.item_link_extractor.extract_links(resp)
        for link in item_links:
            yield Request(link.url, callback=self.parse, priority=20,
                          cb_kwargs={'to_parse': True})

        # Extract all links from this page
        follow_links = self.follow_link_extractor.extract_links(resp)
        for link in follow_links:
            yield SeleniumCallbackRequest(
                link.url, callback=self.parse, priority=10)
    
    def _parse_item(self, resp):
        l = ItemLoader(item=_OikotieApartmentItem(), response=resp)
        l.add_value('url', resp.url)
        l.add_value('time', int(time.time()))
        
        # Apartment info
        l.add_xpath('title', '//title//text()')
        l.add_xpath('overview',
            '//div[contains(@class, "listing-overview")]//text()')

        # From tables
        table_xpath = '//dt[text()="{title}"]/following-sibling::dd[1]//text()'
        for title, field in self.title2field.items():
            l.add_xpath(field, table_xpath.format(title=title))

        # Contact information
        l.add_xpath(
            'contact_person_name',
            '//div[contains(@class, "listing-person__details-item--big")]'
            '//text()'
        )
        l.add_xpath(
            'contact_person_job_title',
            '//div[contains(@class, "listing-person__details-item--waisted")]'
            '//text()'
        )
        l.add_xpath(
            'contact_person_phone_number',
            '(//div[contains(@class, "listing-person__details-item'
            '--sm-top-margin")]/span)[2]//text()'
        )
        l.add_xpath(
            'contact_person_company',
            '//div[@class="listing-company__name"]/a/span//text()'
        )
        l.add_xpath('contact_person_email', '(//p)[1]//text()')
        return l.load_item()


class _OikotieApartmentItem(Item):
    __doc__ = """
    Returned fields:
        * url (str): URL of the scraped web page.
        * time (int): UNIX timestamp of the scraping.
        * title (str): Title of the web browser tab.
        * overview (str): Overview text of the apartment.
        * contact_person_name (str): Name of the contact person.
        * contact_person_job_title (str): Job title of the contact person.
        * contact_person_phone_number (str): Phone number of the contact \
            person.
        * contact_person_company (str): Company of the contact person.
        """.strip() + '\n' + (
            '\n'.join(f'{" " * 8}* {field} (str): {desc}'
            for desc, field in _OikotieApartmentSpider.title2field.items()))
    url = Field(
        input_processor=Identity(),
        output_processor=TakeFirst()
    )
    time = Field(
        input_processor=Identity(),
        output_processor=TakeFirst()
    )
    # Apartment info
    title = Field(
        input_processor=Identity(),
        output_processor=TakeFirst()
    )
    overview = Field(
        input_processor=strip_join,
        output_processor=TakeFirst()
    )
    # Basic information
    location = Field(
        input_processor=strip_join,
        output_processor=TakeFirst()
    )
    city = Field(
        input_processor=Identity(),
        output_processor=TakeFirst()
    )
    house_number = Field(
        input_processor=Identity(),
        output_processor=TakeFirst()
    )
    district = Field(
        input_processor=Identity(),
        output_processor=TakeFirst()
    )
    oikotie_id = Field(
        input_processor=Identity(),
        output_processor=TakeFirst()
    )
    floor = Field(
        input_processor=lambda x: x[-1:],
        output_processor=TakeFirst()
    )
    total_floors = Field(
        input_processor=Identity(),
        output_processor=TakeFirst()
    )
    life_sq = Field(
        input_processor=Identity(),
        output_processor=TakeFirst()
    )
    property_sq = Field(
        input_processor=Identity(),
        output_processor=TakeFirst()
    )
    total_sq = Field(
        input_processor=Identity(),
        output_processor=TakeFirst()
    )
    room_info = Field(
        input_processor=Identity(),
        output_processor=TakeFirst()
    )
    number_of_rooms = Field(
        input_processor=Identity(),
        output_processor=TakeFirst()
    )
    condition = Field(
        input_processor=Identity(),
        output_processor=TakeFirst()
    )
    condition_details = Field(
        input_processor=Identity(),
        output_processor=TakeFirst()
    )
    availability = Field(
        input_processor=Identity(),
        output_processor=TakeFirst()
    )
    kitchen_appliances = Field(
        input_processor=Identity(),
        output_processor=TakeFirst()
    )
    bathroom_appliances = Field(
        input_processor=Identity(),
        output_processor=TakeFirst()
    )
    window_direction = Field(
        input_processor=Identity(),
        output_processor=TakeFirst()
    )
    has_balcony = Field(
        input_processor=Identity(),
        output_processor=TakeFirst()
    )
    balcony_details = Field(
        input_processor=Identity(),
        output_processor=TakeFirst()
    )
    storage_space = Field(
        input_processor=Identity(),
        output_processor=TakeFirst()
    )
    view = Field(
        input_processor=Identity(),
        output_processor=TakeFirst()
    )
    future_renovations = Field(
        input_processor=Identity(),
        output_processor=TakeFirst()
    )
    completed_renovations = Field(
        input_processor=Identity(),
        output_processor=TakeFirst()
    )
    has_sauna = Field(
        input_processor=Identity(),
        output_processor=TakeFirst()
    )
    sauna_details = Field(
        input_processor=Identity(),
        output_processor=TakeFirst()
    )
    housing_type = Field(
        input_processor=Identity(),
        output_processor=TakeFirst()
    )
    services = Field(
        input_processor=Identity(),
        output_processor=TakeFirst()
    )
    additional_info = Field(
        input_processor=Identity(),
        output_processor=TakeFirst()
    )
    property_id = Field(
        input_processor=Identity(),
        output_processor=TakeFirst()
    )
    apartment_is = Field(
        input_processor=Identity(),
        output_processor=TakeFirst()
    )
    telecommunication_services = Field(
        input_processor=Identity(),
        output_processor=TakeFirst()
    )
    # Price and cost information
    price_no_tax = Field(
        input_processor=Identity(),
        output_processor=TakeFirst()
    )
    sales_price = Field(
        input_processor=Identity(),
        output_processor=TakeFirst()
    )
    shared_loan_payment = Field(
        input_processor=Identity(),
        output_processor=TakeFirst()
    )
    price_per_sq = Field(
        input_processor=Identity(),
        output_processor=TakeFirst()
    )
    share_of_liabilities = Field(
        input_processor=Identity(),
        output_processor=TakeFirst()
    )
    mortgages = Field(
        input_processor=Identity(),
        output_processor=TakeFirst()
    )
    financial_charge = Field(
        input_processor=Identity(),
        output_processor=TakeFirst()
    )
    condominium_payment = Field(
        input_processor=Identity(),
        output_processor=TakeFirst()
    )
    maintenance_charge = Field(
        input_processor=Identity(),
        output_processor=TakeFirst()
    )
    water_charge = Field(
        input_processor=Identity(),
        output_processor=TakeFirst()
    )
    water_charge_details = Field(
        input_processor=Identity(),
        output_processor=TakeFirst()
    )
    heating_charge = Field(
        input_processor=Identity(),
        output_processor=TakeFirst()
    )
    heating_charge_details = Field(
        input_processor=Identity(),
        output_processor=TakeFirst()
    )
    other_costs = Field(
        input_processor=Identity(),
        output_processor=TakeFirst()
    )
    # House and property
    is_brand_new = Field(
        input_processor=Identity(),
        output_processor=TakeFirst()
    )
    housing_company_name = Field(
        input_processor=Identity(),
        output_processor=TakeFirst()
    )
    building_type = Field(
        input_processor=Identity(),
        output_processor=TakeFirst()
    )
    build_year = Field(
        input_processor=Identity(),
        output_processor=TakeFirst()
    )
    build_year_details = Field(
        input_processor=Identity(),
        output_processor=TakeFirst()
    )
    number_of_apartments = Field(
        input_processor=Identity(),
        output_processor=TakeFirst()
    )
    building_has_elevator = Field(
        input_processor=Identity(),
        output_processor=TakeFirst()
    )
    building_has_sauna = Field(
        input_processor=Identity(),
        output_processor=TakeFirst()
    )
    building_material = Field(
        input_processor=Identity(),
        output_processor=TakeFirst()
    )
    roof_type = Field(
        input_processor=Identity(),
        output_processor=TakeFirst()
    )
    energy_class = Field(
        input_processor=Identity(),
        output_processor=TakeFirst()
    )
    has_energy_certificate = Field(
        input_processor=Identity(),
        output_processor=TakeFirst()
    )
    antenna_system = Field(
        input_processor=Identity(),
        output_processor=TakeFirst()
    )
    property_size = Field(
        input_processor=Identity(),
        output_processor=TakeFirst()
    )
    property_size_unit = Field(
        input_processor=Identity(),
        output_processor=TakeFirst()
    )
    property_owner = Field(
        input_processor=Identity(),
        output_processor=TakeFirst()
    )
    maintenance = Field(
        input_processor=Identity(),
        output_processor=TakeFirst()
    )
    real_estate_management = Field(
        input_processor=Identity(),
        output_processor=TakeFirst()
    )
    plan_info = Field(
        input_processor=Identity(),
        output_processor=TakeFirst()
    )
    plan = Field(
        input_processor=Identity(),
        output_processor=TakeFirst()
    )
    traffic_communication = Field(
        input_processor=Identity(),
        output_processor=TakeFirst()
    )
    heating = Field(
        input_processor=Identity(),
        output_processor=TakeFirst()
    )
    # Spaces and material
    parking_space_description = Field(
        input_processor=Identity(),
        output_processor=TakeFirst()
    )
    common_spaces = Field(
        input_processor=Identity(),
        output_processor=TakeFirst()
    )
    wallcovering = Field(
        input_processor=Identity(),
        output_processor=TakeFirst()
    )
    # Contacts
    contact_person_name = Field(
        input_processor=strip_join,
        output_processor=TakeFirst()
    )
    contact_person_company = Field(
        input_processor=Identity(),
        output_processor=TakeFirst()
    )
    contact_person_job_title = Field(
        input_processor=Identity(),
        output_processor=TakeFirst()
    )
    contact_person_email = Field(
        input_processor=Identity(),
        output_processor=TakeFirst()
    )
    contact_person_phone_number = Field(
        input_processor=Identity(),
        output_processor=TakeFirst()
    )



"""Module for testing spiders."""


import time

import pytest
pytestmark = [pytest.mark.spider]

from finscraper.spiders import ILArticle, ISArticle, YLEArticle, VauvaPage, \
    OikotieApartment

from tests.utils import calc_field_emptiness


# Spiders can be added here, and basic tests will be set up automatically
spiders = [
    {
        'class': ILArticle,
        'params': [None],
        'n_fields': 8,
        'mark': pytest.mark.ilarticle
    },
    {
        'class': ISArticle,
        'params': [None],
        'n_fields': 8,
        'mark': pytest.mark.isarticle
    },
    {
        'class': YLEArticle,
        'params': [None],
        'n_fields': 8,
        'mark': pytest.mark.ylearticle
    },
    {
        'class': VauvaPage,
        'params': [None],
        'n_fields': 8,
        'mark': pytest.mark.vauvapage
    },
    {
        'class': OikotieApartment,
        'params': [None],
        'n_fields': 80,
        'mark': pytest.mark.oikotieapartment
    }
]


scrape_cases = [pytest.param(s['class'], p, s['n_fields'], marks=s['mark'])
                for s in spiders for p in s['params']]

other_cases = [pytest.param(s['class'], p, marks=s['mark'])
               for s in spiders for p in s['params']]         


@pytest.fixture(scope='function')
def spider_params(request):
    return request.param or {}


@pytest.mark.parametrize('spider_cls, spider_params, n_fields', scrape_cases,
                         indirect=['spider_params'])
def test_scraping(spider_cls, spider_params, n_fields, capsys, n_items=10):
    # Scraping
    spider = spider_cls(**spider_params).scrape(n_items)
    df = spider.get()
    assert len(df) >= n_items
    assert len(df.columns) == n_fields

    # Continue scraping
    df2 = spider.scrape(1).get()
    assert len(df2) > len(df)

    # Field emptiness
    emptiness = calc_field_emptiness(df2)
    with capsys.disabled():
        print(f'\n{emptiness.to_string(index=False)}\n')
    assert not (emptiness['empty_pct'] == 100).all()


@pytest.mark.parametrize('spider_cls, spider_params', other_cases,
                         indirect=['spider_params'])
def test_functionality(spider_cls, spider_params):
    # Save and load
    spider = spider_cls(**spider_params).scrape(1)
    df = spider.get()
    jobdir = spider.save()
    del spider
    spider = spider_cls.load(jobdir)
    df2 = spider.scrape(1).get()
    assert len(df2) > len(df)


@pytest.mark.parametrize('spider_cls, spider_params', other_cases,
                         indirect=['spider_params'])
@pytest.mark.xfail(reason='Benchmark')
def test_benchmark_scrape_1_min(spider_cls, spider_params, capsys):
    df = spider_cls(**spider_params).scrape(0, 60).get()
    assert len(df) >= 60
    with capsys.disabled():
        print(f'-- {len(df)} items')


@pytest.mark.parametrize('spider_cls, spider_params', other_cases,
                         indirect=['spider_params'])
@pytest.mark.xfail(reason='Benchmark')
def test_benchmark_scrape_100_items(spider_cls, spider_params, capsys):
    start = time.perf_counter()
    df = spider_cls(**spider_params).scrape(100).get()
    elapsed_time = int(time.perf_counter() - start)
    assert len(df) >= 100
    with capsys.disabled():
        print(f'-- {elapsed_time} seconds ({len(df)} items)')



"""Module for utility functions and classes."""


import io
import logging
import pickle
import re

# Monkey patch, see https://github.com/pypa/pipenv/issues/2609
import webdriver_manager.utils
def console(text, bold=False):
    pass
webdriver_manager.utils.console = console

from selenium import webdriver
from selenium.webdriver.chrome.options import Options
from webdriver_manager.chrome import ChromeDriverManager

from tqdm.auto import tqdm


class TqdmLogger(io.StringIO):
    """File-like object that redirects buffer to stdout."""
    def __init__(self, logger):
        self.logger = logger
        self.buf = ''

    def write(self, buf):
        self.buf = buf

    def flush(self):
        if self.buf.strip() != '':
            self.logger.log(logging.DEBUG, self.buf)


class QueueHandler(logging.Handler):
    """Sends events to a queue, allowing multiprocessing.

    This handler checks for picklability before saving items into queue.
    Modified from: https://gist.github.com/vsajip/591589
    """
    def __init__(self, queue): 
        logging.Handler.__init__(self)
        self.queue = queue

    def _get_picklable_attrs(self, record):
        # TODO: More performant way to do the same
        attrdict = {}
        for attr in vars(record):
            value = getattr(record, attr)
            try:
                pickle.dumps(value)
                attrdict[attr] = value
            except AttributeError:
                pass
            except:
                pass
        
        if type(record.args) == tuple:
            attrdict['args'] = record.args
        else:
            args = {}
            for attr, value in record.args.items():
                try:
                    pickle.dumps(value)
                    args[attr] = value
                except AttributeError:
                    args[attr] = str(value)
                    pass
                except:
                    pass
            attrdict['args'] = args
        new_record = logging.makeLogRecord(attrdict)
        return new_record

    def enqueue(self, record):
        self.queue.put_nowait(record)

    def prepare(self, record):
        record = self._get_picklable_attrs(record)
        self.format(record)
        record.msg = record.message
        record.args = None
        record.exc_info = None
        return record

    def emit(self, record):
        try:
            self.enqueue(self.prepare(record))
        except (KeyboardInterrupt, SystemExit):
            raise
        except:
            self.handleError(record)


def get_chromedriver(options=None):
    if not options:
        options = Options()
        options.add_argument("--headless")
        options.add_argument("--disable-extensions")
        options.add_argument("--disable-gpu")
    driver = webdriver.Chrome(ChromeDriverManager().install(), options=options)
    return driver


def strip_join(text_list, join_with=' '):
    joined_text = join_with.join(text.strip() for text in text_list
                                 if text is not None)
    return joined_text


def strip_elements(text_list):
    return [text.strip() for text in text_list if text is not None]


def drop_empty_elements(text_list):
    return [text for text in text_list
            if text.strip() != '' and text is not None]

def safe_cast_int(text):
    try:
        return int(text)
    except:
        return None



"""Module for custom Scrapy HTTP components."""


from scrapy import Request


class SeleniumCallbackRequest(Request):
    """Process request with given callback using Selenium.

    Args:
        selenium_callback (func or None): Function that will be called with the
            chrome webdriver. The function should take in parameters 
            (request, spider, driver) and return request, response or None.
            If None, driver will be used for fetching the page, and return is
            response. Defaults to None.
    """

    def __init__(self, *args, selenium_callback=None, **kwargs):
        #if not callable(selenium_callback):
        #    raise ValueError('`callback` must be a function!')
        meta = kwargs.pop('meta', {})
        if 'selenium_callback' not in meta:
            meta['selenium_callback'] = selenium_callback
        new_kwargs = dict(**kwargs, meta=meta)
        super(SeleniumCallbackRequest, self).__init__(*args, **new_kwargs)



"""Module for ISArticle spider."""


import time

from scrapy import Item, Field, Selector
from scrapy.crawler import Spider
from scrapy.linkextractors import LinkExtractor
from scrapy.loader import ItemLoader
from scrapy.loader.processors import TakeFirst, Identity, MapCompose

from finscraper.scrapy_spiders.mixins import FollowAndParseItemMixin
from finscraper.utils import strip_join


class _ISArticleSpider(FollowAndParseItemMixin, Spider):
    name = 'isarticle'
    start_urls = ['https://www.is.fi']
    follow_link_extractor = LinkExtractor(
        allow_domains=('is.fi'),
        allow=(),
        deny=(r'.*/tag/.*', r'.*/haku/.*', r'.*/reseptit/.*', r'.*/mainos/.*',
              r'.*/yritys/.*'),
        deny_domains=('ravit.is.fi'),
        canonicalize=True
    )
    item_link_extractor = LinkExtractor(
        allow_domains=('is.fi'),
        allow=(rf'.*/art\-[0-9]+\.html'),
        deny=(r'.*/tag/.*', r'.*/haku/.*', r'.*/reseptit/.*', r'.*/mainos/.*',
              r'.*/yritys/.*'),
        deny_domains=('ravit.is.fi'),
        canonicalize=True
    )
    def __init__(self, *args, **kwargs):
        """Fetch IltaSanomat news articles.
        
        Args:
        """
        super(_ISArticleSpider, self).__init__(*args, **kwargs)

    @staticmethod
    def _get_image_metadata(text):
        sel = Selector(text=text)
        return {
            'src': sel.xpath('//img//@src').get(),
            'alt': sel.xpath('//img//@alt').get(),
            'caption': strip_join(sel.xpath('//p//text()').getall())
        }
    
    def _parse_item(self, resp):
        l = ItemLoader(item=_ISArticleItem(), response=resp)
        l.add_value('url', resp.url)
        l.add_value('time', int(time.time()))
        l.add_xpath('title', '//article//h1//text()')
        l.add_xpath('ingress',
            '//section//article//p[contains(@class, "ingress")]//text()')
        l.add_xpath('content',
            '//article//p[contains(@class, "body")]//text()')
        l.add_xpath('published',
            '//article//div[contains(@class, "timestamp")]//text()')
        l.add_xpath('author',
            '//article//div[contains(@itemprop, "author")]//text()')
        l.add_xpath('images',
            '//section//article//div[contains(@class, "clearing-container")]')
        return l.load_item()


class _ISArticleItem(Item):
    """
    Returned fields:
        * url (str): URL of the scraped web page.
        * time (int): UNIX timestamp of the scraping.
        * title (str): Title of the article.
        * ingress (str): Ingress of the article.
        * content (str): Contents of the article.
        * published (str): Publish time of the article.
        * author (str): Author of the article.
        * images (list of dict): Images of the article.
    """
    url = Field(
        input_processor=Identity(),
        output_processor=TakeFirst()
    )
    time = Field(
        input_processor=Identity(),
        output_processor=TakeFirst()
    )
    title = Field(
        input_processor=strip_join,
        output_processor=TakeFirst()
    )
    ingress = Field(
        input_processor=strip_join,
        output_processor=TakeFirst()
    )
    content = Field(
        input_processor=strip_join,
        output_processor=TakeFirst()
    )
    published = Field(
        input_processor=strip_join,
        output_processor=TakeFirst()
    )
    author = Field(
        input_processor=strip_join,
        output_processor=TakeFirst()
    )
    images = Field(
        input_processor=MapCompose(_ISArticleSpider._get_image_metadata),
        output_processor=Identity()
    )



"""Module for VauvaPage spider."""


import time

from functools import partial

from scrapy import Item, Field, Selector
from scrapy.crawler import Spider
from scrapy.exceptions import DropItem
from scrapy.linkextractors import LinkExtractor
from scrapy.loader import ItemLoader
from scrapy.loader.processors import TakeFirst, Identity, MapCompose, Compose

from finscraper.scrapy_spiders.mixins import FollowAndParseItemMixin
from finscraper.utils import strip_join, safe_cast_int, strip_elements, \
    drop_empty_elements


class _VauvaPageSpider(FollowAndParseItemMixin, Spider):
    name = 'vauvapage'
    start_urls = ['https://vauva.fi']
    follow_link_extractor = LinkExtractor(
        allow_domains=('vauva.fi'),
        allow=(r'.*\/keskustelu\/.*'),
        deny=('quote', 'changed', 'user', r'(\&|\?)(rate)\='),
        deny_domains=(),
        canonicalize=True
    )
    item_link_extractor = LinkExtractor(
        allow_domains=('vauva.fi'),
        allow=(rf'keskustelu/[0-9]+/[A-z0-9\-]+'),
        deny=('quote', 'changed', 'user', r'(\&|\?)(rate)\='),
        deny_domains=(),
        canonicalize=True
    )
    custom_settings = {}

    def __init__(self, *args, **kwargs):
        """Fetch comments from vauva.fi.
        
        Args:
        """
        super(_VauvaPageSpider, self).__init__(*args, **kwargs)

    def _parse_comment(self, comment):
        l = ItemLoader(item=_VauvaCommentItem(), selector=comment)
        l.add_xpath('author', '//*[contains(@property, "name")]//text()')
        l.add_xpath('date', '//div[contains(@class, "post-date")]//text()')
        l.add_xpath('quotes', '//blockquote//text()', strip_join)
        l.add_xpath('content', '//p//text()')
        votes = l.nested_xpath('//span[contains(@class, "voting-count")]')
        votes.add_xpath('upvotes', '//li[@class="first"]//text()')
        votes.add_xpath('downvotes', '//li[@class="last"]//text()')
        return l.load_item()
    
    def _parse_item(self, resp):
        l = ItemLoader(item=_VauvaPageItem(), response=resp)
        l.add_value('url', resp.url)
        l.add_value('time', int(time.time()))
        l.add_xpath('title',
            '//article//div[contains(@property, "title")]//text()')
        l.add_xpath('page',
            '//article//li[contains(@class, "pager-current")]//text()')
        l.add_value('page', ['1'])
        l.add_xpath('pages',
            '//article//li[contains(@class, "pager-last")]//text()')
        l.add_xpath('pages',
            '//article//li[contains(@class, "pager-current")]//text()')
        l.add_value('pages', ['1'])
        l.add_xpath('published',
            '(//article//div[contains(@class, "post-date")])[1]//text()')
        l.add_xpath('author',
            '(//article//*[contains(@property, "name")])[1]//text()')
        
        comments = []
        comment_xpath = '//article//article[contains(@class, "comment")]'
        for comment in resp.xpath(comment_xpath):
            comments.append(self._parse_comment(Selector(text=comment.get())))
        l.add_value('comments', comments)
        loaded_item = l.load_item()
        return loaded_item


class _VauvaCommentItem(Item):
    """
    Returned comment fields:
        * author (str): Author of the comment.
        * date (str): Publish time of the comment.
        * quotes (list of str): List of quotes in the comment.
        * content (str): Contents of the comment.
        * upvotes (int): Upvotes of the comment.
        * downvotes (int): Downvotes of the comment.
    """
    author = Field(
        input_processor=strip_join,
        output_processor=TakeFirst()
    )
    date = Field(
        input_processor=strip_join,
        output_processor=Compose(strip_elements, TakeFirst())
    )
    quotes = Field(
        input_processor=drop_empty_elements,
        output_processor=Identity()
    )
    content = Field(
        input_processor=strip_join,
        output_processor=TakeFirst()
    )
    upvotes = Field(
        input_processor=MapCompose(TakeFirst(), safe_cast_int),
        output_processor=TakeFirst()
    )
    downvotes = Field(
        input_processor=MapCompose(TakeFirst(), safe_cast_int),
        output_processor=TakeFirst()
    )


class _VauvaPageItem(Item):
    __doc__ = """
    Returned page fields:
        * url (str): URL of the scraped web page.
        * time (int): UNIX timestamp of the scraping.
        * title (str): Title of the thread.
        * page (int): Page number of the thread.
        * pages (int): Pages in the thread.
        * comments (str): Comments of the thread page.
        * published (str): Publish time of the article.
        * author (str): Author of the article.
    """ + _VauvaCommentItem.__doc__
    url = Field(
        input_processor=Identity(),
        output_processor=TakeFirst()
    )
    time = Field(
        input_processor=Identity(),
        output_processor=TakeFirst()
    )
    title = Field(
        input_processor=strip_join,
        output_processor=TakeFirst()
    )
    page = Field(
        input_processor=MapCompose(safe_cast_int),
        output_processor=TakeFirst()
    )
    pages = Field(
        input_processor=MapCompose(safe_cast_int),
        output_processor=TakeFirst()
    )
    comments = Field(
        input_processor=Identity(),
        output_processor=Identity()
    )
    published = Field(
        input_processor=strip_join,
        output_processor=Compose(strip_elements, TakeFirst())
    )
    author = Field(
        input_processor=strip_join,
        output_processor=TakeFirst()
    )



"""Module for Spider API - the main interface of finscraper."""


from textwrap import indent

from finscraper.wrappers import _SpiderWrapper
from finscraper.scrapy_spiders.ilarticle import _ILArticleSpider, \
    _ILArticleItem
from finscraper.scrapy_spiders.isarticle import _ISArticleSpider, \
    _ISArticleItem
from finscraper.scrapy_spiders.vauvapage import _VauvaPageSpider, \
    _VauvaPageItem, _VauvaPageSpider
from finscraper.scrapy_spiders.ylearticle import _YLEArticleSpider, \
    _YLEArticleItem
from finscraper.scrapy_spiders.oikotieapartment import \
    _OikotieApartmentSpider, _OikotieApartmentItem


__wrapper_doc__ = '''
jobdir (str or None, optional): Working directory of the spider.
    Defaults to None, which creates a temp directory to be used.
    Note that this directory will only be deleted through the ``clear`` method!
progress_bar (bool, optional): Whether to enable progress bar or not. This
    parameter is ignored if ``log_level`` is not None. Defaults to True.
log_level (str or None, optional): Logging level to display. Should be in
    ['debug', 'info', 'warn', 'error', 'critical'] or None (disabled).
    Defaults to None.
    
    .. note::
        This parameter can be overridden through Scrapy ``settings``
        (LOG_LEVEL, LOG_ENABLED) within the ``scrape`` -method.
'''


def _get_docstring(spider_cls, item_cls):
    return (spider_cls.__init__.__doc__.strip()
            + indent(__wrapper_doc__, ' ' * 12)
            + indent(item_cls.__doc__, ' ' * 4))


class ISArticle(_SpiderWrapper):
    __doc__ = _get_docstring(_ISArticleSpider, _ISArticleItem)
    def __init__(self, jobdir=None, progress_bar=True, log_level=None):
        super(ISArticle, self).__init__(
            spider_cls=_ISArticleSpider,
            spider_params=dict(),
            jobdir=jobdir,
            progress_bar=progress_bar,
            log_level=log_level)


class ILArticle(_SpiderWrapper):
    __doc__ = _get_docstring(_ILArticleSpider, _ILArticleItem)
    def __init__(self, jobdir=None, progress_bar=True, log_level=None):
        super(ILArticle, self).__init__(
            spider_cls=_ILArticleSpider,
            spider_params=dict(),
            jobdir=jobdir,
            progress_bar=progress_bar,
            log_level=log_level)


class YLEArticle(_SpiderWrapper):
    __doc__ = _get_docstring(_YLEArticleSpider, _YLEArticleItem)
    def __init__(self, jobdir=None, progress_bar=True, log_level=None):
        super(YLEArticle, self).__init__(
            spider_cls=_YLEArticleSpider,
            spider_params=dict(),
            jobdir=jobdir,
            progress_bar=progress_bar,
            log_level=log_level)


class VauvaPage(_SpiderWrapper):
    __doc__ = _get_docstring(_VauvaPageSpider, _VauvaPageItem)
    def __init__(self, jobdir=None, progress_bar=True, log_level=None):
        super(VauvaPage, self).__init__(
            spider_cls=_VauvaPageSpider,
            spider_params=dict(),
            jobdir=jobdir,
            progress_bar=progress_bar,
            log_level=log_level)


class OikotieApartment(_SpiderWrapper):
    __doc__ = _get_docstring(_OikotieApartmentSpider, _OikotieApartmentItem)
    def __init__(self, jobdir=None, progress_bar=True, log_level=None):
        super(OikotieApartment, self).__init__(
            spider_cls=_OikotieApartmentSpider,
            spider_params=dict(),
            jobdir=jobdir,
            progress_bar=progress_bar,
            log_level=log_level)



"""Module for Scrapy middlewares."""


import logging

import time

from scrapy import signals, Request
from scrapy.exceptions import NotConfigured
from scrapy.http import HtmlResponse

from selenium.webdriver.chrome.options import Options

from finscraper.http import SeleniumCallbackRequest
from finscraper.utils import get_chromedriver


class SeleniumCallbackMiddleware:
    """Middleware that processes request with given callback."""

    def __init__(self, settings):
        self.settings = settings

    @classmethod
    def from_crawler(cls, crawler):
        middleware = cls(crawler.settings)
        crawler.signals.connect(middleware.spider_opened,
                                signals.spider_opened)
        crawler.signals.connect(middleware.spider_closed,
                                signals.spider_closed)
        return middleware

    def spider_opened(self, spider):
        options = Options()
        options.add_argument("--headless")
        options.add_argument("--disable-extensions")
        options.add_argument("--disable-gpu")
        if self.settings.get('PROGRESS_BAR_ENABLED', True):
            options.add_argument('--disable-logging')
            for name in ['selenium.webdriver.remote.remote_connection',
                         'requests', 'urllib3']:
                logging.getLogger(name).propagate = False
        self.driver = get_chromedriver(options)

    def spider_closed(self, spider):
        self.driver.close()

    def process_request(self, request, spider):
        if not isinstance(request, SeleniumCallbackRequest):
            return None
        
        selenium_callback = request.meta.get('selenium_callback')
        if selenium_callback is None:
            self.driver.get(request.url)
            return HtmlResponse(
                self.driver.current_url,
                body=self.driver.page_source.encode('utf-8'),
                encoding='utf-8',
                request=request
            )
        else:
            return selenium_callback(request, spider, self.driver)



"""Module for Scrapy extensions."""


import logging

from collections import defaultdict

from scrapy import signals
from scrapy.exceptions import NotConfigured, CloseSpider

from tqdm.auto import tqdm

from finscraper.utils import TqdmLogger


class ProgressBar:
    """Scrapy extension thay displays progress bar.
    
    Enabled via ``PROGRESS_BAR_ENABLED`` Scrapy setting.
    """
    def __init__(self, crawler):
        self.progress_bar_enabled = (
            crawler.settings.get('PROGRESS_BAR_ENABLED', False))
        self.closespider_itemcount = (
            crawler.settings.get('CLOSESPIDER_ITEMCOUNT', None))
        self.counter = defaultdict(int)

        if not self.progress_bar_enabled:
            raise NotConfigured

        crawler.signals.connect(self.on_response,
                                signal=signals.response_received)
        crawler.signals.connect(self.on_item_scraped,
                                signal=signals.item_scraped)
        crawler.signals.connect(self.on_error, signal=signals.spider_error)

        logger = logging.getLogger()
        self.progress_bar = tqdm(desc='Progress', unit=' items',
                                 total=self.closespider_itemcount,
                                 file=TqdmLogger(logger))
        self.itemcount = 0
        self.pagecount = 0

    @classmethod
    def from_crawler(cls, crawler):
        return cls(crawler)

    def on_error(self, failure, response, spider):
        self.counter['errorcount'] += 1
        self.progress_bar.set_postfix({
            'pages': self.counter['pagecount'],
            'errors': self.counter['errorcount']
        })

    def on_response(self, response, request, spider):
        self.counter['pagecount'] += 1
        self.progress_bar.set_postfix({
            'pages': self.counter['pagecount'],
            'errors': self.counter['errorcount']
        })

    def on_item_scraped(self, item, spider):
        self.counter['itemcount'] += 1
        self.progress_bar.update()



"""Module for wrapping Scrapy spiders."""


import json
import logging
import multiprocessing as mp
import pickle
import platform
import shutil
import sys
import tempfile
import uuid
import weakref

from logging.handlers import QueueListener
from pathlib import Path

import pandas as pd

from scrapy import Request
from scrapy.crawler import CrawlerProcess, CrawlerRunner
from scrapy.exceptions import CloseSpider
from scrapy.settings import Settings
from scrapy.spiders import Spider
from scrapy.utils.log import configure_logging

from twisted.internet import reactor

from finscraper.utils import QueueHandler


# https://turtlemonvh.github.io/python-multiprocessing-and-corefoundation-libraries.html
if platform.system() == 'Darwin':
    mp = mp.get_context('spawn')


def _run_as_process(func, spider_cls, spider_params, settings):
    # Setup logging / progress bar
    # (queuehandler --> listener --> root logger --> streamhandler)
    progress_bar_enabled = settings['PROGRESS_BAR_ENABLED']
    log_enabled = settings['LOG_ENABLED']
    log_stdout = settings['LOG_STDOUT']
    q_log = None
    ql = None
    if log_enabled or progress_bar_enabled:
        handler = logging.StreamHandler()
        if progress_bar_enabled:
            handler.terminator = ''
            handler.setFormatter(logging.Formatter('%(message)s'))
        else:
            handler.setFormatter(logging.Formatter(settings.get('LOG_FORMAT')))

        q_log = mp.Queue(-1)
        ql = QueueListener(q_log, handler)
        ql.start()

        logger = logging.getLogger()
        logger.setLevel(settings.get('LOG_LEVEL'))
        logger.addHandler(handler)
    
    # Start function as a separate process
    q = mp.Queue()
    p = mp.Process(target=func,
                   args=(q, q_log, spider_cls, spider_params, settings))
    p.start()
    result = q.get()
    p.join()

    if ql:
        ql.stop()

    if isinstance(result, BaseException):
        raise result


def _run_spider_func(q, q_log, spider_cls, spider_params, settings):
    try:
        configure_logging(settings, install_root_handler=False)
        if q_log is not None:
            # Setup logging (worker --> queuehandler --> root logger)
            if not settings['LOG_ENABLED']:  # Disables STDOUT :o
                logging.getLogger('scrapy').propagate = False
            qh = QueueHandler(q_log)
            logger = logging.getLogger()
            logger.setLevel(settings.get('LOG_LEVEL'))
            qh.setLevel(settings.get('LOG_LEVEL'))
            qh.setFormatter(logging.Formatter(settings.get('LOG_FORMAT')))
            logger.addHandler(qh)

        # Start crawling
        runner = CrawlerRunner(settings)
        deferred = runner.crawl(spider_cls, **spider_params)
        deferred.addBoth(lambda _: reactor.stop())
        reactor.run()
        q.put(None)
    except Exception as e:
        q.put(e)


class _SpiderWrapper:
    """Provide common methods and attributes for all spiders."""
    _log_levels = {
        'debug': logging.DEBUG,
        'info': logging.INFO,
        'warn': logging.WARN,
        'error': logging.ERROR,
        'critical': logging.CRITICAL
    }

    def __init__(self, spider_cls, spider_params, jobdir=None,
                 progress_bar=True, log_level=None):
        self.spider_cls = spider_cls
        self.spider_params = spider_params

        if jobdir is None:
            self._jobdir = Path(tempfile.gettempdir()) / str(uuid.uuid4())
        elif type(jobdir) == str:
            self._jobdir = Path(jobdir)
        else:
            raise ValueError(f'Jobdir {jobdir} is not of type str or None')
        self._jobdir.mkdir(parents=True, exist_ok=True)

        self.log_level = log_level
        self.progress_bar = progress_bar and self.log_level is None
        
        self._items_save_path = self._jobdir / 'items.jl'
        self._spider_save_path = self._jobdir / 'spider.pkl'

        self._finalizer = weakref.finalize(
            self, shutil.rmtree, self._jobdir, ignore_errors=True)
        
        # Note: Parameters cannot be changed outside by setting them
        for param in self.spider_params:
            setattr(self, param, self.spider_params[param])

    @property
    def jobdir(self):
        """Working directory of the spider.
        
        Can be changed after initialization of a spider.
        """
        return str(self._jobdir)

    @property
    def log_level(self):
        """Logging level of the spider.
        
        This attribute can be changed after initialization of a spider.
        """
        return self._log_level

    @log_level.setter
    def log_level(self, log_level):
        if log_level is None:
            self._log_level = log_level
        elif (type(log_level) == str 
              and log_level.strip().lower() in self._log_levels):
            self._log_level = self._log_levels[log_level.strip().lower()]
        else:
            raise ValueError(
                    f'Log level should be in {self._log_levels.keys()}')

    @property
    def progress_bar(self):
        """Whether progress bar is enabled or not.
        
        Can be changed after initialization of a spider.
        """
        return self._progress_bar

    @progress_bar.setter
    def progress_bar(self, progress_bar):
        if type(progress_bar) == bool:
            self._progress_bar = progress_bar
        else:
            raise ValueError(f'Progress bar "{progress_bar}" not boolean')

    @property
    def items_save_path(self):
        """Save of path of the scraped items.
        
        Cannot be changed after initialization of a spider.
        """
        return str(self._items_save_path)

    @property
    def spider_save_path(self):
        """Save path of the spider.
        
        Cannot be changed after initialization of a spider.
        """
        return str(self._spider_save_path)

    def _run_spider(self, itemcount=10, timeout=60, pagecount=0, errorcount=0,
                    settings=None):
        _settings = Settings()
        _settings.setmodule('finscraper.settings', priority='project')
        
        _settings['JOBDIR'] = self.jobdir
        _settings['FEEDS'] = {self.items_save_path: {'format': 'jsonlines'}}
        
        _settings['CLOSESPIDER_ITEMCOUNT'] = itemcount
        _settings['CLOSESPIDER_TIMEOUT'] = timeout
        _settings['CLOSESPIDER_PAGECOUNT'] = pagecount
        _settings['CLOSESPIDER_ERRORCOUNT'] = errorcount
        
        _settings['LOG_STDOUT'] = True
        _settings['LOG_LEVEL'] = self.log_level or logging.NOTSET
        _settings['LOG_ENABLED'] = self.log_level is not None
        # Logging dominates progress bar
        _settings['PROGRESS_BAR_ENABLED'] = self.progress_bar
        
        # Will always be prioritized --> conflicts are possible
        if settings is not None:
            _settings.update(settings)
        
        try:
            _run_as_process(
                func=_run_spider_func,
                spider_cls=self.spider_cls,
                spider_params=self.spider_params,
                settings=_settings
            )
        except KeyboardInterrupt:
            pass
    
    def scrape(self, n=10, timeout=60, settings=None):
        """Scrape given number of items.
        
        Args:
            n (int, optional): Number of items to attempt to scrape. Zero
                corresponds to no limit. Defaults to 10.
            timeout (int, optional): Timeout in seconds to wait before stopping
                the spider. Zero corresponds to no limit. Defaults to 60.
            settings (dict or None, optional): Scrapy spider settings to use.
                Defaults to None, which correspond to default settings.
                See list of available settings at:
                https://docs.scrapy.org/en/latest/topics/settings.html.
        
        Returns:
            self
        """
        self._run_spider(itemcount=n, timeout=timeout, settings=settings)
        return self

    def get(self, fmt='df'):
        """Return scraped data as DataFrame or list.
        
        Args:
            fmt (str, optional): Format to return parsed items as. Should be
                in ['df', 'list']. Defaults to 'df'.

        Returns:
            If ``fmt = 'df'``, DataFrame of scraped items.
            If ``fmt = 'list'``, list of dict of scraped items.

        Raises:
            ValueError: If ``fmt`` not in allowed values.
        """
        if fmt not in ['df', 'list']:
            ValueError(f'Format {fmt} should be in ["df", "list"]')
        jsonlines = []
        if self._items_save_path.exists():
            with open(self.items_save_path, 'r') as f:
                for line in f:
                    jsonlines.append(json.loads(line))
        if fmt == 'list':
            return jsonlines
        elif fmt == 'df':
            return pd.DataFrame(jsonlines)
    
    def save(self):
        """Save spider in ``jobdir`` for later use.
        
        Returns:
            str: Path to job directory.
        """
        save_tuple = (self.spider_cls, self.spider_params, self.jobdir)
        with open(self.spider_save_path, 'wb') as f:
            pickle.dump(save_tuple, f)
        self._finalizer.detach()
        return self.jobdir

    @classmethod
    def load(cls, jobdir):
        """Load existing spider from ``jobdir``.
        
        Args:
            jobdir (str): Path to job directory.
        
        Returns:
            Spider loaded from job directory.
        """
        expected_path = Path(jobdir) / 'spider.pkl'
        with open(expected_path, 'rb') as f:
            (spider_cls, spider_params, jobdir) = pickle.load(f)
        return cls(jobdir=jobdir, **spider_params)

    def clear(self):
        """Clear contents of ``jobdir``."""
        if self._jobdir.exists():
            shutil.rmtree(self._jobdir)
        self._jobdir.mkdir(parents=True, exist_ok=True)



"""Module for Scrapy spider mixins."""


from scrapy import Request
from scrapy.exceptions import CloseSpider


class FollowAndParseItemMixin:
    """Parse items and follow links based on defined link extractors.
    
    The following needs to be defined when inheriting:
        1) ``item_link_extractor`` -attribute: LinkExtractor that defines \
            the links to parse items from.
        2) ``follow_link_extractor`` -attribute: LinkExtractor that defines \
            the links to follow and find item pages from.
        3) ``parse_item`` -function: Parses the item from response.

    Args:
        follow_meta (dict or None, optional): Dictionary to pass within \
            link follow requests. Defaults to None.
        follow_items (dict or None, optional): Dictionary to pass within \
            item link requests. Defaults to None.

    Raises:
        AttributeError, if required attributes not defined when inheriting.
    """
    itemcount = 0
    def __init__(self, follow_meta=None, items_meta=None):
        self.follow_meta = follow_meta
        self.items_meta = items_meta

        for attr in ['follow_link_extractor', 'item_link_extractor']:
            if not hasattr(self, attr):
                raise AttributeError(f'Attribute `{attr}` must be defined')

    def _parse_item(self, resp):
        raise NotImplementedError('Function `_parse_item` not implemented!')

    def start_requests(self):
        for url in self.start_urls:
            yield Request(url, callback=self.parse, meta=self.follow_meta)

    def parse(self, resp, to_parse=False):
        """Parse items and follow links based on defined link extractors."""
        if (self.itemcount and 
            self.itemcount == self.settings.get('CLOSESPIDER_ITEMCOUNT', 0)):
                raise CloseSpider

        if to_parse:
            yield self._parse_item(resp)
            self.itemcount += 1

        # Parse items and further on extract links from those pages
        item_links = self.item_link_extractor.extract_links(resp)
        for link in item_links:
            yield Request(link.url, callback=self.parse, meta=self.items_meta,
                          priority=20, cb_kwargs={'to_parse': True})

        # Extract all links from this page
        follow_links = self.follow_link_extractor.extract_links(resp)
        for link in follow_links:
            yield Request(link.url, callback=self.parse, meta=self.follow_meta,
                          priority=10)



"""Module for Scrapy pipelines."""


class DefaultValueNonePipeline:
    """Pipeline that sets default values of all item fields to None."""

    def process_item(self, item, spider):
        for field in item.fields:
            item.setdefault(field, None)
        return item



"""Module for testing spider wrapper functionalities."""


import logging

from pathlib import Path

from finscraper.spiders import ISArticle, ILArticle


def test_spider_save_load_with_jobdir():
    jobdir = '../jobdir'
    spider = ISArticle(jobdir=jobdir)
    
    save_jobdir = spider.save()
    loaded_spider = ISArticle.load(save_jobdir)

    assert (jobdir
            == str(spider.jobdir)
            == save_jobdir == str(loaded_spider.jobdir))


def test_spider_save_load_without_jobdir():
    spider = ISArticle()
    
    save_jobdir = spider.save()
    loaded_spider = ISArticle.load(save_jobdir)

    assert spider.jobdir is not None
    assert (str(spider.jobdir) == save_jobdir == str(loaded_spider.jobdir))


def test_spider_clear():
    # Directory removed not saved and deleted
    spider = ISArticle()
    jobdir = spider.jobdir
    del spider
    assert not Path(jobdir).exists()

    # Directory not removed if saved and deleted
    spider = ISArticle().scrape(1)
    jobdir = spider.save()
    del spider
    assert Path(jobdir).exists()

    # Directory contents removed when cleared
    spider = ISArticle().scrape(1)
    jobdir = spider.save()
    spider_save_path = spider.spider_save_path
    items_save_path = spider.items_save_path
    assert Path(spider_save_path).exists()
    assert Path(items_save_path).exists()
    spider.clear()
    assert not Path(spider_save_path).exists()
    assert not Path(items_save_path).exists()


def test_spider_logging():
    # Yes attribute "None", no settings
    spider = ISArticle(log_level=None)
    spider.scrape(1)

    # Yes attribute, no settings
    spider = ISArticle(log_level='info')
    spider.scrape(1)
    assert True

    # No attribute, yes settings
    spider = ISArticle(log_level=None)
    spider.scrape(1, settings={'LOG_LEVEL': logging.DEBUG})
    assert True

    spider = ISArticle(log_level='info')
    spider.scrape(1, settings={'LOG_ENABLED': False})
    assert True

    # Attribute set
    try:
        spider.log_level = 'test'
    except ValueError:
        assert True
    except:
        assert False
    spider.log_level = 'info'
    assert spider.log_level == logging.INFO

    # TODO: Test the output


def test_spider_progress_bar():
    # Progress bar true by default
    spider = ILArticle()
    spider.scrape(1)
    assert spider.progress_bar == True
    assert len(spider.get()) > 0

    # Progress bar disabled, when log level given
    spider = ILArticle(log_level='info')
    spider.scrape(1)
    assert spider.progress_bar == False
    assert len(spider.get()) > 0

    # TODO: Test the output
